/* $Id: parser.mly,v 1.25 2001/11/18 23:15:43 malekith Exp $ 
 *
 * Parser in ocamlyacc for the gont language.
 */
%{
open Gontutil
;;

let locus () = {loc_file=Lexer.get_file_name (); 
                loc_char=Parsing.symbol_start ()}

let ice () = raise Parse_error

let make_exp raw =
  {raw_exp = raw;
   exp_ty = None;
   exp_loc = locus ()}

let bin op e1 e2 =
  make_exp (SimpleBinOp(op, e1, e2))
  
let make_stmt raw =
  {raw_stmt = raw;
   stmt_loc = locus ()}
   
let make_ref id =
  make_exp (Ref {ref_name = id; 
                 ref_level = -1;
                 ref_symbol = None})
let make_pair a b =
  make_stmt (Pair (a, b))
;;

let unary op e =
  {raw_exp = SimpleUnOp(op, e);
   exp_ty = None;
   exp_loc = locus ()}
;;

%}

%token <string> INT
%token <string> FLOAT
%token <string> ID
%token <string> STRING
%token <string> TYVAR           /* 'a -> TYVAR("a") */
%token <char> CHAR
%token KW_INT KW_RETURN KW_VOID KW_SKIP KW_BOOL KW_FLOAT 
%token KW_EXTERN_C KW_STRING KW_FUN KW_IF KW_ELSE KW_WHILE KW_DO KW_FOR
%token KW_CASE KW_SWITCH KW_LET KW_EXPORT_AS
%token COLON DOUBLECOLON ARROW LSQUARE RSQUARE
%token LT GT GE LE EQ NE LTLT GTGT
%token PLUS MINUS TIMES DIV EQUAL MOD
%token PLUS_DOT MINUS_DOT TIMES_DOT DIV_DOT
%token LPAREN RPAREN LBRACE RBRACE SEMICOLON COMMA

/* lowest precedence */
%right EQUAL
%left LT GT GE LE
%left EQ NE
%left PLUS MINUS PLUS_DOT MINUS_DOT
%left TIMES DIV TIMES_DOT DIV_DOT MOD
%nonassoc UMINUS
%nonassoc APPLICATION
/* highest precedence */

%start decl iface_decl            /* the entry points */

%type <Gontutil.decl> decl
%type <Gontutil.decl> iface_decl
%type <Gontutil.ty> type_exp

%%

decl:
          func          { $1 }
        | global_var    { $1 }
        | extern_c      { $1 }
;

iface_decl:
          func_proto    { $1 }
        | global_var    { $1 }
;

id: 
        ID              { $1 }
;

maybe_export_as:
         /* empty */            { "" }
       | KW_EXPORT_AS STRING    { $2 }
;

func:
       type_exp id LPAREN func_args RPAREN maybe_export_as lambda_body
                { FuncDecl {fn_loc = locus ();
                            fn_ret_type = $1; 
                            fn_name = $2;
                            fn_parms = $4;
                            fn_body = $7;
                            fn_symbol = None;
                            fn_export_as = $6} }
;

global_var:
       type_exp id SEMICOLON
                { GlobalDecl {global_loc = locus ();
                              global_type = $1;
                              global_name = $2;
                              global_symbol = None} }
;

/* from now on ne_ == non_empty_ */

func_arg:
        type_exp id             { {parm_type = $1; 
                                   parm_name = $2; 
                                   parm_symbol = None} }
;

ne_func_args:
          func_arg                      { [$1] }
        | func_arg COMMA ne_func_args   { $1 :: $3 }
;

func_args:
          /* empty */    { [] }
        | ne_func_args   { $1 }
;

proto_arg:
          type_exp              { $1 }
        | type_exp id           { $1 }
;

ne_proto_args:
          proto_arg                             { [$1] }
        | proto_arg COMMA ne_proto_args         { $1 :: $3 }
;

proto_args:
          /* empty */           { [] }
        | ne_proto_args         { $1 }
;

proto:
        type_exp id LPAREN proto_args RPAREN
                { ($1, $2, $4) }
;

ne_expr_list:
          expr                           { [$1] }
        | expr COMMA ne_expr_list        { $1 :: $3 }
;

expr_list:
          /* empty */    { [] }
        | ne_expr_list   { $1 }
;

extern_c:
          KW_EXTERN_C proto SEMICOLON {
            let (t, n, a) = $2 in
              ExternC {xc_type = Func(t, a);
                       xc_name = n;
                       xc_loc = locus ();
                       xc_symbol = None} }
        | KW_EXTERN_C type_exp id SEMICOLON {
              ExternC {xc_type = $2;
                       xc_name = $3;
                       xc_loc = locus ();
                       xc_symbol = None} }
;

func_proto:
        proto SEMICOLON {
            let (t, n, a) = $1 in
              FuncProto {fp_type = Func(t, a);
                         fp_name = n;
                         fp_loc = locus ();
                         fp_symbol = None} }
;

block:
        LBRACE stmts RBRACE { 
          make_stmt (match $2 with
                        [] -> Skip
                      | _ -> Block $2) 
        }
;

stmts:
          /* empty */      { [] }
        | stmt stmts       { $1::$2 }
;

stmt:
          normal_stmt SEMICOLON         { $1 }
        | block                         { $1 }
        | func_def                      { $1 }
        | control_stmt                  { $1 }
        | pattern_matching_stmt         { $1 }
;

  
var_def:
          type_exp id           { make_stmt (VarDef {var_type = $1; 
                                                     var_name = $2; 
                                                     var_symbol = None}) }
        | type_exp id EQUAL expr { 
             let def =  make_stmt (VarDef {var_type = $1; 
                                           var_name = $2; 
                                           var_symbol = None}) in
             let init = make_stmt (Exp (bin Assign (make_ref $2) $4)) in
             make_pair def init
        }
;

func_def:
        type_exp id LPAREN func_args RPAREN lambda_body {
           let ty = Func($1, List.map (fun {parm_type = t} -> t) $4) in
           let body = make_exp (Lambda {lambda_ret_type = $1;
                                        lambda_parms = $4;
                                        lambda_body = $6}) in
           let def = make_stmt (VarDef {var_type = ty; 
                                        var_name = $2; 
                                        var_symbol = None}) in
           let init = bin Assign (make_ref $2) body in
           make_pair def (make_stmt (Exp init))
        }
;

for_exp1:
          stmt_expr             { make_stmt (Exp $1) }
        | var_def               { $1 }
        | KW_SKIP               { make_stmt Skip }
        | /* empty */           { make_stmt Skip }
;

for_exp2:
          expr                  { $1 }
        | /* empty */           { make_exp (BoolConst true) }
;

for_exp3:
          stmt_expr             { make_stmt (Exp $1) }
        | KW_SKIP               { make_stmt Skip }
        | /* empty */           { make_stmt Skip }
;        

control_stmt:
          KW_IF LPAREN expr RPAREN stmt { 
            make_stmt (If ($3, $5, make_stmt Skip)) 
        }
        | KW_IF LPAREN expr RPAREN block KW_ELSE block {
            make_stmt (If ($3, $5, $7)) 
        }
        | KW_WHILE LPAREN expr RPAREN stmt {
              let cond = make_stmt (BreakIfNot $3) in
              make_stmt(Loop(make_pair cond $5)) 
        }
        | KW_DO block KW_WHILE LPAREN expr RPAREN {
              let cond = make_stmt (BreakIfNot $5) in
              make_stmt(Loop(make_pair $2 cond)) 
        }
        | KW_FOR LPAREN 
                for_exp1 SEMICOLON 
                for_exp2 SEMICOLON 
                for_exp3 RPAREN stmt {
              let cond = make_stmt (BreakIfNot $5) in
              let loop = Loop(make_pair cond (make_pair $9 $7)) in
              make_stmt (Block [$3; make_stmt loop])
        }
;

normal_stmt:
          stmt_expr             { make_stmt (Exp $1) }
        | KW_RETURN             { make_stmt ReturnVoid }
        | KW_RETURN expr        { make_stmt (Return $2) }
        | KW_SKIP               { make_stmt Skip }
        | var_def               { $1 }
;

variable:
        id                      { make_ref $1 }
;

tuple_content:
          expr COMMA expr               { [$1; $3] }
        | expr COMMA tuple_content      { $1 :: $3 }
;

tuple:
        LSQUARE tuple_content RSQUARE     { ice () }
;

/* expression that might be use to form expression stmt */
stmt_expr:
          call %prec APPLICATION  { $1 }
        | expr EQUAL expr         { bin Assign $1 $3 }
;

expr:
          primary                 { $1 }
        | stmt_expr               { $1 }
	| expr PLUS expr          { bin IntPlus $1 $3 }
	| expr MINUS expr         { bin IntMinus $1 $3 }
	| expr TIMES expr         { bin IntMult $1 $3 }
	| expr DIV expr           { bin IntDiv $1 $3 }
	| expr MOD expr           { bin IntMod $1 $3 }
        | expr PLUS_DOT expr      { bin FloatPlus $1 $3 }
	| expr MINUS_DOT expr     { bin FloatMinus $1 $3 }
	| expr TIMES_DOT expr     { bin FloatMult $1 $3 }
	| expr DIV_DOT expr       { bin FloatDiv $1 $3 }
        | expr LT expr            { bin Lt $1 $3 }
        | expr GT expr            { bin Gt $1 $3 }
        | expr GE expr            { bin Ge $1 $3 }
        | expr LE expr            { bin Le $1 $3 }
        | expr EQ expr            { bin Eq $1 $3 }
        | expr NE expr            { bin Ne $1 $3 }
	| MINUS expr %prec UMINUS { unary IntNegate $2 }
	| MINUS_DOT expr %prec UMINUS { unary FloatNegate $2 }
;

primary:
          INT                     { make_exp (IntConst $1) }
        | FLOAT                   { make_exp (FloatConst $1) }
        | STRING                  { make_exp (StringConst $1) }
        | variable                { $1 }
	| LPAREN expr RPAREN      { $2 }
        | lambda                  { $1 }
        | tuple                   { $1 }
        | id LSQUARE expr_list RSQUARE          { ice () }
;

call:
        primary LPAREN expr_list RPAREN  { make_exp (Call($1, $3)) }
;
        
simple_type:
	  KW_VOID	{ Void }
	| KW_INT	{ Int }
        | KW_BOOL       { Bool }
        | KW_FLOAT      { Float }
        | KW_STRING     { String }
;

lambda_header:
           KW_FUN LPAREN func_args RPAREN COLON type_exp ARROW  
                { ($3, $6) }
;

lambda_body:
          block { $1 }
        | LPAREN expr RPAREN {
                let e = make_stmt (Return $2) in
                make_stmt (Block([e]))
        }
;

lambda:
          lambda_header lambda_body {
                let (p, r) = $1 in
                make_exp (Lambda {lambda_ret_type = r;
                                  lambda_parms = p;
                                  lambda_body = $2}) }
;

function_type:
      TIMES LPAREN proto_args RPAREN ARROW type_exp     { Func($6, $3) }
;

tuple_ty:
          type_exp COMMA ne_ty_list     { $1 :: $3 }
;

ne_ty_list:
          type_exp                      { [$1] }
        | type_exp COMMA ne_ty_list     { $1 :: $3 }
;


type_exp: 
	  simple_type                   { $1 }
        | function_type                 { $1 }
        | TYVAR                         { TyVar($1) }
        | id                            { ice () }
        | LT ne_ty_list GT id           { ice () }
        /* this is hack to allow <<t1>x>y */
        | LTLT ne_ty_list GT ne_ty_list GT id { ice () }
        | TIMES LSQUARE tuple_ty RSQUARE        { ice () }
;


ne_pattern_list:
          pattern COMMA ne_pattern_list   { $1 :: $3 }
        | pattern COMMA pattern           { [$1; $3] }
;

pattern_list:
          /* empty */                     { [] }
        | ne_pattern_list                 { $1 }
        | pattern                         { [$1] }
;

pattern:
          id                                    { ice() }
        | id LSQUARE pattern_list RSQUARE       { ice() }
        | LSQUARE ne_pattern_list RSQUARE       { ice() }
;


case_clouse:
        KW_CASE pattern COLON stmts     { ice() }
;

case_list:
          case_clouse case_list           { $1 :: $2 }
        | /* empty */                     { [] }
;

pattern_matching_stmt:
          KW_LET pattern EQUAL expr block         { ice() }
        | KW_SWITCH expr LBRACE case_list RBRACE  { ice() }
;
